#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  3 01:08:10 2019

"""
import numpy as np
from AD.DualNumber import DualNumber
import AD.ElementaryFunctions as EF

class Parallelized_AD:
    """
    DESCRIPTION
    =======
    A class to run automatic differentiation in both forward and reverse mode for vector
    functions of vectors and scalars.  It stores a function or a list of functions, where
    each function is a string (similar to Matlab function handles).  A list of variables 
    is required to calculate the derivative.  For instance, 'sin(x1)' must be written as 
    func ='sin(_x1)' and var_names = 'x1'.  To initalize a Parallelized_AD instance, the user must then
    write PAD=Parallelized_AD(fun=func,var=var_names).
    
    INPUTS
    =======
    fun: array, default None
        Stores the list of functions for which we will eventually calculate the Jacobian. Note that all variables
        in the function array must be preceded by an underscore '_', e.g. _x.
    var: array, default None
        Stores list of variables (will differentiate function wrt each)
    Reverse: Boolean, optional, if set to true the gradient will be calculated based on reverse mode
        Default value is false, in which case forward mode is used.
    RETURNS
    ========
    Parallelized_AD: An instance with methods such as get_Jacobian() to get the value of the Jacobian
    at a certain point.  The user can also add functions to the current list of functions with the 
    add_function() method.
    
    
    NOTES
    =====
    PRE:
        - fun, var are arrays or strings
        - get_Jacobian() method takes numeric loc array as an input, the location where 
        the Jacobian should be evaluated
        
    POST:
        - function, variables will be stored in public attributes self.function, self.variable
        - self.Jacobian is initialized to None
        - raises a AttributeError exception if function input is not a list or string
        - raises a AttributeError exception if variable input is not a list or string
        - defines the overloading of arithmetic operators and unary operators
        
    EXAMPLES
    =========
    >>> Parallelized_AD(fun=['_x**_y + logistic(_x+_y) + _z'],var=['x','y', 'z']).get_Jacobian([0.5,0.25,1])
    array([[ 0.6383432 , -0.36496999,  1.        ]])
    >>> func = ['_x + sin(_y)*_z', '_x + sin(_y)*exp(_z)']
    >>> PAD = Parallelized_AD(fun = func, var = ['x', 'y', 'z'])
    >>> PAD.get_Jacobian([1,2,3])
    array([[ 1.        , -1.24844051,  0.90929743],
           [ 1.        , -8.35853265, 18.26372704]])
    >>> func = ['_x + sin(_y)*_z', '_x + sin(_y)*exp(_z)']
    >>> PAD = Parallelized_AD(fun = func, var = ['x', 'y', 'z'])
    >>> PAD.add_function('_x+_y+1')
    >>> PAD.get_Jacobian([1,2,3], forward=True)
    array([[ 1.        , -1.24844051,  0.90929743],
           [ 1.        , -8.35853265, 18.26372704],
           [ 1.        ,  1.        ,  0.        ]])
    """

    def __init__(self,fun=None,var=None):
        if fun:
            assert isinstance(fun,(str,list))
            self.function=fun if isinstance(fun,list) else [fun]
        else:
            self.function=[]
        if var:
            assert isinstance(var,(str,list))
            self.varname=var if isinstance(var,list) else [var]
        else:
            self.varname=[]
        self.variable=[]
        self._Jacobian=None
        self._value = None
    
    @property
    def Jacobian(self):
        return self._Jacobian
    
    @property
    def value(self):
        return self._value
        
    def add_function(self,fun):
        """
        DESCRIPTION
        =======
        A class method to add a function to an existing Parallelized_AD object.  For example:
        >>> PAD = Parallelized_AD(fun=['_x**_y + sin(_x) + _z'],var=['x','y', 'z'])
        >>> PAD.add_function('_x')
        >>> print(PAD.function)
        ['_x**_y + sin(_x) + _z', '_x']
        
        Assert conditions ensure input is a string
        """
        assert isinstance(fun,str)
        self.function=[self.function] if isinstance(self.function,str) else self.function
        self.function.append(fun)
            
    def add_var(self,var):
        """
        DESCRIPTION
        =======
        A class method to add variables to an existing Parallelized_AD object.  For example:
        >>> PAD = Parallelized_AD(fun=['_x**_y + sin(_x)'],var=['x','y'])
        >>> PAD.add_function('_z')
        >>> PAD.add_var('z')
        >>> print(PAD.function)
        ['_x**_y + sin(_x)', '_z']
        >>> print(PAD.varname)
        ['x', 'y', 'z']
        
        Assert conditions ensure input is a string
        """
        assert isinstance(var,str)
        self.varname=[self.varname] if isinstance(self.varname,str) else self.varname
        self.varname.append(var)
            
    def get_Jacobian(self,loc,forward=False):
        """
        DESCRIPTION
        =======
        A class method to get the Jacobian of user-specified vector function.  See
        doctests for details on usage.  User inputs location to take Jacobian, and
        function returns the Jacobian at that point.  Assert conditions ensure 
        dimension of location input is the same as the number of variables.
        
        forward argument specifies whether reverse (False) or forward (True) mode
        is used
        """

        assert len(loc)==len(self.varname) 
        self._Jacobian=np.zeros((len(self.function),len(self.varname)))
        
        # for each function, if forward, do forward mode calculation, else do reverse
        # see documentation for details on reverse mode calculation
        for i, fun in enumerate(self.function):
            if forward:
                # pre-process each function to be differentatied
                translated_fun=self.preprocess(fun)
                # for each variable, take the derivative at the value specified
                for j in range(len(self.varname)):
                    self.variable=[DualNumber(value,dual=0) for value in loc]   
                    self.variable[j]=DualNumber(loc[j],dual=1) 
                    element=eval(translated_fun)
                    self._Jacobian[i,j]=element.der
            # similarly, if reverse, carry out reverse mode
            else:
                
                self.variable=[DualNumber(value,Reverse=True) for value in loc]
                translated_fun=self.preprocess(fun)
                element=eval(translated_fun)
                element.set_der(1)
                for j in range(len(self.varname)):
                    self._Jacobian[i,j]=self.variable[j].der
        return self._Jacobian
    
    def get_value(self,loc):
        """
        DESCRIPTION
        =======
        A class method to get the value of user-specified vector function.  See
        doctests for details on usage.  User inputs location of vector-valued function,
        and get_value returns the value at the specified location.
        
        >>> PAD = Parallelized_AD(fun=['_x * arcSin(_y*_z)+_x'], var=['x', 'y', 'z'])
        >>> print(PAD.get_value([0.4,0.2,1]))
        [0.48054317]
        >>> print(PAD.get_Jacobian([0.4,0.2,1]))
        [[1.20135792 0.40824829 0.08164966]]
        """
        assert len(loc)==len(self.varname)
        self._value=np.zeros((len(self.function)))
        
        # for each function, if forward, do forward mode calculation, else do reverse
        # see documentation for details on reverse mode calculation
        for i, fun in enumerate(self.function):
            
            # pre-process each function to be differentatied
            translated_fun=self.preprocess(fun)

            # for each variable, take the derivative at the value specified
            # for each variable, take the derivative at the value specified
            self.variable=[DualNumber(value) for value in loc]
            element=eval(translated_fun)
            self._value[i]=element.val
        return self._value
        
    
    def preprocess(self, string:str):
        """
        DESCRIPTION
        =======
        A class method to process the user input functions into a form such that we 
        can use our DualNumber class and ElementaryFunction module.
        
        >>> PAD = Parallelized_AD(fun=['sin(_x)'],var=['x'])
        >>> print(PAD.preprocess(PAD.function[0]))
        EF.Sin(self.variable[0])
        """
        dictionary={'sin(':'EF.Sin(',
                    'tan(':'EF.Tan(',
                    'cos(':'EF.Cos(',
                    'exp(':'EF.Exp(',
                    'power(':'EF.Power(',
                    'log(':'EF.Log(',
                    'arcSin(':'EF.ArcSin(',
                    'arcCos(':'EF.ArcCos(',
                    'arcTan(':'EF.ArcTan(',
                    'sqrt(':'EF.Sqrt(',
                    'sinh(':'EF.Sinh(',
                    'cosh(':'EF.Cosh(',
                    'tanh(':'EF.Tanh(',
                    'logistic(':'EF.Logistic('}
        for key,item in dictionary.items():
            string=string.replace(key,item)    
        for i,name in enumerate(self.varname):
            string=string.replace('_' + name,f'self.variable[{i}]')

        return string
   
if __name__=='__main__':   
    
    import doctest
    doctest.testmod(verbose=True)
    



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-

from AD.DualNumber import DualNumber
import AD.ElementaryFunctions as EF
from AD.Parallelized import Parallelized_AD
import numpy as np
import pytest

# DualNumber tests
def test_overload_add():
    x = DualNumber(5,1)
    y = DualNumber(7,1)

    assert (x + y).val == 12 and (x + y).der == 2
    assert (x + 5).val == 10 and (x + 5).der == 1
    assert (x + 5.0).val == 10 and (x + 5.0).der == 1
    assert (5 + x).val == 10 and (5 + x).der == 1
    assert (5.0 + x).val == 10 and (5.0 + x).der == 1

def test_overload_add_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x+"test"

def test_overload_multiply():
    x = DualNumber(5)
    y = DualNumber(7,2)

    assert (x * y).val == 35 and (x * y).der == 17
    assert (x * 5).val == 25 and (x * 5).der == 5
    assert (x * 5.0).val == 25 and (x * 5.0).der == 5
    assert (5.0 * x).val == 25 and (5.0 * x).der == 5
    assert (5 * x).val == 25 and (5 * x).der == 5

def test_overload_multiply_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x*"test"

def test_overload_sub():
    x = DualNumber(5,2)
    y = DualNumber(7,1)

    assert (x - y).val == -2 and (x - y).der == 1
    assert (x.__rsub__(y)).val == 2 and (x.__rsub__(y)).der == -1
    assert (x - 5).val == 0 and (x - 5).der == 2
    assert (x - 5.0).val == 0 and (x - 5.0).der == 2
    assert (5 - x).val == 0 and (5 - x).der == -2
    assert (5.0 - x).val == 0 and (5.0 - x).der == -2

def test_overload_sub_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x - "test"

def test_overload_truediv():
    x = DualNumber(10,2)
    y = DualNumber(2,4)
    
    assert (x / y).val == 5 and (x / y).der == (2*2-10*4)/(2**2)
    assert (x.__rtruediv__(y)).val == 0.2 and (x.__rtruediv__(y)).der == (4*10-2*2)/(10*10)
    assert (x / 5).val == 2 and (x / 5).der == 0.4
    assert (x / 5.0).val == 2.0 and (x / 5.0).der == 0.4
    assert (5 / x).val == 0.5 and (5 / x).der == -0.1
    assert (5.0 / x).val == 0.5 and (5.0 / x).der == -0.1

def test_overload_truediv_types():
    with pytest.raises(AssertionError):
      x = DualNumber(10)
      x/"test"

def test_overload_power():
    x = DualNumber(2,2)
    y = DualNumber(3,4)

    assert (x**y).val == 8 and (x**y).der == 8*(3/2*2+4*np.log(2))
    assert (x.__rpow__(y)).val == 9 and (x.__rpow__(y)).der == 9*(2/3*4+2*np.log(3))
    assert (x**2).val == 4 and (x**2).der == 8
    assert (x**2.0).val == 4 and (x**2.0).der == 8
    assert (2**x).val == 4 and (2**x).der == np.log(2)*4*2
    assert (2.0**x).val == 4 and (2.0**x).der == np.log(2)*4*2

def test_overload_power_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x**"test"

def test_overload_pos():
    x = DualNumber(5,1)
    y = +x
    assert y.val == 5 and y.der == 1

def test_overload_neg():
    x = DualNumber(5,1)
    y = -x
    assert y.val == -5 and y.der == -1

def test_overload_abs():
    x = DualNumber(-5,-1)
    y = abs(x)
    assert y.val == 5 and y.der == 1
    x = DualNumber(5,1)
    y = abs(x)
    assert y.val == 5 and y.der == 1

def test_overload_round():
    x = DualNumber(5.123,1.0012)
    y = round(x,2)
    assert y.val == 5.12 and y.der == 1.00
    
def test_repr():
    x = DualNumber(5,1)
    assert repr(x) == 'Derivative: 1.00' + '\n' + 'Value: 5.00'
    
def test_str():
    x = DualNumber(5,1)
    assert print(x) == print('Derivative: 1.00' + '\n' + 'Value: 5.00')
    
def test_set_der():
    x = DualNumber(5,1)
    x.set_der(2)
    assert x.der == 2

def test_set_val():
    x = DualNumber(5,1)
    x.set_val(2)
    assert x.val == 2
    
def test_jacobian():
    x = DualNumber(5,1)
    assert x.jacobian == 1

def test_overload_eq():
    x = DualNumber(5,1)
    y = DualNumber(5,1)
    z = DualNumber(6,1)
    assert x == y
    assert not x == z

def test_overload_ne():
    w = DualNumber(5,1)
    x = DualNumber(5,1)
    y = DualNumber(6,1)
    z = DualNumber(5,2)
    assert not w != x
    assert x != y
    assert x != z
    
def test_overload_lt():
    x = DualNumber(5,1)
    y = DualNumber(6,1)
    assert x < y
    assert not y < x
    assert x < 6

def test_overload_le():
    x = DualNumber(5,1)
    y = DualNumber(6,1)
    assert x <= y
    assert not y <= x
    assert x <= 6
    
def test_overload_gt():
    x = DualNumber(6,1)
    y = DualNumber(5,1)
    assert x > y
    assert not y > x
    assert x > 5

def test_overload_ge():
    x = DualNumber(6,1)
    y = DualNumber(5,1)
    assert x >= y
    assert not y >= x
    assert x >= 5
    
test_overload_add()
test_overload_add_types()
test_overload_multiply()
test_overload_multiply_types()
test_overload_sub()
test_overload_sub_types()
test_overload_truediv()
test_overload_truediv_types()
test_overload_power()
test_overload_power_types()
test_overload_pos()
test_overload_neg()
test_overload_abs()
test_overload_round()
test_repr()
test_str()
test_set_der()
test_set_val()
test_jacobian()
test_overload_eq()
test_overload_ne()
test_overload_lt()
test_overload_le()
test_overload_gt()
test_overload_ge()

# Elementary Function tests
def test_sin():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Sin(Test_Dual_Number_1).val == np.sin(1) and EF.Sin(Test_Dual_Number_1).der == np.cos(1)
    Test_Dual_Number_1 = 1
    assert EF.Sin(Test_Dual_Number_1).val == np.sin(1) and EF.Sin(Test_Dual_Number_1).der == 0

def test_cos():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Cos(Test_Dual_Number_1).val == np.cos(1) and EF.Cos(Test_Dual_Number_1).der == -np.sin(1)
    Test_Dual_Number_1 = 1
    assert EF.Cos(Test_Dual_Number_1).val == np.cos(1) and EF.Cos(Test_Dual_Number_1).der == 0

def test_tan():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Tan(Test_Dual_Number_1).val == np.tan(1) and EF.Tan(Test_Dual_Number_1).der == 1 / np.cos(1)**2
    Test_Dual_Number_1 = 1
    assert EF.Tan(Test_Dual_Number_1).val == np.tan(1) and EF.Tan(Test_Dual_Number_1).der == 0

def test_exp():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Exp(Test_Dual_Number_1).val == np.exp(1) and EF.Exp(Test_Dual_Number_1).der == np.exp(1)
    Test_Dual_Number_1 = 1
    assert EF.Exp(Test_Dual_Number_1).val == np.exp(1) and EF.Exp(Test_Dual_Number_1).der == 0

def test_Power():
    Test_Dual_Number_2 = DualNumber(2)
    assert EF.Power(Test_Dual_Number_2, 2).val == 4 and EF.Power(Test_Dual_Number_2, 2).der == 4
    Test_Dual_Number_2 = 2
    assert EF.Power(Test_Dual_Number_2, 2).val == 4 and EF.Power(Test_Dual_Number_2, 2).der == 0

def test_Log():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Log(Test_Dual_Number_1).val == 0 and EF.Log(Test_Dual_Number_1).der == 1
    Test_Dual_Number_1 = 1
    assert EF.Log(Test_Dual_Number_1).val == 0 and EF.Log(Test_Dual_Number_1).der == 0
    
    Test_Dual_Number_1 = DualNumber(2)
    assert EF.Log(Test_Dual_Number_1,base=10).val == np.log(2) / np.log(10)  and EF.Log(Test_Dual_Number_1,base=10).der == 1/(np.log(10)*2)
    Test_Dual_Number_1 = 2
    assert EF.Log(Test_Dual_Number_1,base=10).val == np.log(2) / np.log(10) and EF.Log(Test_Dual_Number_1,base=10).der == 0

def test_ArcSin():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcSin(Test_Dual_Number_1).val == np.arcsin(0.5) and EF.ArcSin(Test_Dual_Number_1).der == 1 / np.sqrt(
        0.75)
    Test_Dual_Number_1 = 0.5
    assert EF.ArcSin(Test_Dual_Number_1).val == np.arcsin(0.5) and EF.ArcSin(Test_Dual_Number_1).der == 0
    with pytest.raises(ValueError):
        x = DualNumber(5)
        EF.ArcSin(x)
        
def test_ArcCos():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcCos(Test_Dual_Number_1).val == np.arccos(0.5) and EF.ArcCos(Test_Dual_Number_1).der == -1 / np.sqrt(
        0.75)
    Test_Dual_Number_1 = 0.5
    assert EF.ArcCos(Test_Dual_Number_1).val == np.arccos(0.5) and EF.ArcCos(Test_Dual_Number_1).der == 0
    with pytest.raises(ValueError):
        x = DualNumber(5)
        EF.ArcCos(x)

def test_ArcTan():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcTan(Test_Dual_Number_1).val == np.arctan(0.5) and EF.ArcTan(Test_Dual_Number_1).der == 1 / 1.25
    Test_Dual_Number_1 = 0.5
    assert EF.ArcTan(Test_Dual_Number_1).val == np.arctan(0.5) and EF.ArcTan(Test_Dual_Number_1).der == 0

def test_Sqrt():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.Sqrt(Test_Dual_Number_1).val == np.sqrt(0.5) and EF.Sqrt(Test_Dual_Number_1).der == 1 / np.sqrt(
        0.5) / 2
    Test_Dual_Number_1 = 0.5
    assert EF.Sqrt(Test_Dual_Number_1).val == np.sqrt(0.5) and EF.Sqrt(Test_Dual_Number_1).der == 0


def test_sinh():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Sinh(Test_Dual_Number_1).val == np.sinh(1) and EF.Sinh(Test_Dual_Number_1).der == np.cosh(1)
    Test_Dual_Number_1 = 1
    assert EF.Sinh(Test_Dual_Number_1).val == np.sinh(1) and EF.Sin(Test_Dual_Number_1).der == 0

def test_cosh():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Cosh(Test_Dual_Number_1).val == np.cosh(1) and EF.Cosh(Test_Dual_Number_1).der == np.sinh(1)
    Test_Dual_Number_1 = 1
    assert EF.Cosh(Test_Dual_Number_1).val == np.cosh(1) and EF.Cosh(Test_Dual_Number_1).der == 0

def test_tanh():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Tanh(Test_Dual_Number_1).val == np.tanh(1) and round(EF.Tanh(Test_Dual_Number_1).der,8) == round((1 / np.cosh(1)**2),8)
    Test_Dual_Number_1 = 1
    assert EF.Tanh(Test_Dual_Number_1).val == np.tanh(1) and EF.Tanh(Test_Dual_Number_1).der == 0

def test_Logistic():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Logistic(Test_Dual_Number_1).val == 1 / (1 + np.exp(-1)) and round(EF.Logistic(Test_Dual_Number_1).der,8) == round(np.exp(-1) / (1+np.exp(-1))**2,8)
    Test_Dual_Number_1 = 1
    assert EF.Logistic(Test_Dual_Number_1).val == 1 / (1 + np.exp(-1)) and EF.Logistic(Test_Dual_Number_1).der == 0

def test_data_type_check():
    x = 'x'
    try:
        EF.data_type_check(x)
    except AttributeError:
        return 0
    
test_sin()      
test_cos()
test_tan()
test_exp()
test_Power()
test_Log()
test_ArcSin()
test_ArcCos()
test_ArcTan()
test_Sqrt()
test_sinh()      
test_cosh()
test_tanh()
test_Logistic()
test_data_type_check()

# Reverse Mode tests
def rm_test_overload_add():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(7,Reverse=True)

    assert (x + y).val == 12 and (x + y).der == 0
    assert (x + 5).val == 10 and (x + 5).der == 0
    assert (x + 5.0).val == 10 and (x + 5.0).der == 0
    assert (5 + x).val == 10 and (5 + x).der == 0
    assert (5.0 + x).val == 10 and (5.0 + x).der == 0

def rm_test_overload_add_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5,Reverse=True)
        x+"test"

def rm_test_overload_multiply():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(7,Reverse=True)

    assert (x * y).val == 35 and (x * y).der == 0
    assert (x * 5).val == 25 and (x * 5).der == 0
    assert (x * 5.0).val == 25 and (x * 5.0).der == 0
    assert (5.0 * x).val == 25 and (5.0 * x).der == 0
    assert (5 * x).val == 25 and (5 * x).der == 0

def rm_test_overload_multiply_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5,Reverse=True)
        x*"test"

def rm_test_overload_sub():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(7,Reverse=True)

    assert (x - y).val == -2 and (x - y).der == 0
    assert (x.__rsub__(y)).val == 2 and (x.__rsub__(y)).der == 0
    assert (x - 5).val == 0 and (x - 5).der == 0
    assert (x - 5.0).val == 0 and (x - 5.0).der == 0
    assert (5 - x).val == 0 and (5 - x).der == 0
    assert (5.0 - x).val == 0 and (5.0 - x).der == 0

def rm_test_overload_sub_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5,Reverse=True)
        x - "test"

def rm_test_overload_truediv():
    x = DualNumber(10,Reverse=True)
    y = DualNumber(2,Reverse=True)

    assert (x / y).val == 5 and (x / y).der == 0
    assert (x.__rtruediv__(y)).val == 0.2 and (x.__rtruediv__(y)).der == 0
    assert (x / 5).val == 2 and (x / 5).der == 0
    assert (x / 5.0).val == 2.0 and (x / 5.0).der == 0
    assert (5 / x).val == 0.5 and (5 / x).der == 0
    assert (5.0 / x).val == 0.5 and (5.0 / x).der == 0

def rm_test_overload_truediv_types():
    with pytest.raises(AssertionError):
      x = DualNumber(10,Reverse=True)
      x/"test"

def rm_test_overload_power():
    x = DualNumber(2,Reverse=True)
    y = DualNumber(3,Reverse=True)

    assert (x**y).val == 8 and (x**y).der == 0
    assert (x.__rpow__(y)).val == 9 and (x.__rpow__(y)).der == 0
    assert (x**2).val == 4 and (x**2).der == 0
    assert (x**2.0).val == 4 and (x**2.0).der == 0
    assert (2**x).val == 4 and (2**x).der == 0
    assert (2.0**x).val == 4 and (2.0**x).der == 0

def rm_test_overload_power_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5,Reverse=True)
        x**"test"

def rm_test_overload_pos():
    x = DualNumber(5,Reverse=True)
    y = +x
    assert y.val == 5 and y.der == 0

def rm_test_overload_neg():
    x = DualNumber(5,Reverse=True)
    y = -x
    assert y.val == -5 and y.der == 0

def rm_test_overload_abs():
    x = DualNumber(-5,Reverse=True)
    y = abs(x)
    assert y.val == 5 and y.der == 0
    x = DualNumber(5,1,Reverse=True)
    y = abs(x)
    assert y.val == 5 and y.der == 0

def rm_test_overload_round():
    x = DualNumber(5.123,Reverse=True)
    y = round(x,2)
    assert y.val == 5.12 and y.der == 0
    
def rm_test_repr():
    x = DualNumber(5,Reverse=True)
    assert repr(x) == 'Derivative: 0.00' + '\n' + 'Value: 5.00'
    
def rm_test_str():
    x = DualNumber(5,Reverse=True)
    assert print(x) == print('Derivative: 0.00' + '\n' + 'Value: 5.00')
    
def rm_test_set_der():
    x = DualNumber(5,Reverse=True)
    x.set_der(2)
    assert x.der == 2

def rm_test_set_val():
    x = DualNumber(5,Reverse=True)
    x.set_val(2)
    assert x.val == 2
    
def rm_test_jacobian():
    x = DualNumber(5,Reverse=True)
    assert x.jacobian == 0

def rm_test_overload_eq():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(5,Reverse=True)
    assert x == y

def rm_test_overload_ne():
    w = DualNumber(5,Reverse=True)
    x = DualNumber(5,Reverse=True)
    y = DualNumber(6,Reverse=True)
    assert not w!= x
    assert x != y
    
def rm_test_overload_lt():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(6,Reverse=True)
    assert x < y
    assert not y < x
    assert x < 6

def rm_test_overload_le():
    x = DualNumber(5,Reverse=True)
    y = DualNumber(6,Reverse=True)
    assert x <= y
    assert not y <= x
    assert x <= 6
    
def rm_test_overload_gt():
    x = DualNumber(6,Reverse=True)
    y = DualNumber(5,Reverse=True)
    assert x > y
    assert not y > x
    assert x > 5

def rm_test_overload_ge():
    x = DualNumber(6,Reverse=True)
    y = DualNumber(5,Reverse=True)
    assert x >= y
    assert not y >= x
    assert x >= 5

def rm_test_sin():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Sin(Test_Dual_Number_1).val == np.sin(1) and EF.Sin(Test_Dual_Number_1).der == 0

def rm_test_cos():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Cos(Test_Dual_Number_1).val == np.cos(1) and EF.Cos(Test_Dual_Number_1).der == 0

def rm_test_tan():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Tan(Test_Dual_Number_1).val == np.tan(1) and EF.Tan(Test_Dual_Number_1).der == 0

def rm_test_exp():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Exp(Test_Dual_Number_1).val == np.exp(1) and EF.Exp(Test_Dual_Number_1).der == 0

def rm_test_Power():
    Test_Dual_Number_2 = DualNumber(2,Reverse=True)
    assert EF.Power(Test_Dual_Number_2, 2).val == 4 and EF.Power(Test_Dual_Number_2, 2).der == 0

def rm_test_Log():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Log(Test_Dual_Number_1).val == 0 and EF.Log(Test_Dual_Number_1).der == 0
    
    Test_Dual_Number_1 = DualNumber(2,Reverse=True)
    assert EF.Log(Test_Dual_Number_1,base=10).val == np.log(2) / np.log(10)  and EF.Log(Test_Dual_Number_1,base=10).der == 0

def rm_test_ArcSin():
    Test_Dual_Number_1 = DualNumber(0.5,Reverse=True)
    assert EF.ArcSin(Test_Dual_Number_1).val == np.arcsin(0.5) and EF.ArcSin(Test_Dual_Number_1).der == 0

def rm_test_ArcCos():
    Test_Dual_Number_1 = DualNumber(0.5,Reverse=True)
    assert EF.ArcCos(Test_Dual_Number_1).val == np.arccos(0.5) and EF.ArcCos(Test_Dual_Number_1).der == 0

def rm_test_ArcTan():
    Test_Dual_Number_1 = DualNumber(0.5,Reverse=True)
    assert EF.ArcTan(Test_Dual_Number_1).val == np.arctan(0.5) and EF.ArcTan(Test_Dual_Number_1).der == 0

def rm_test_Sqrt():
    Test_Dual_Number_1 = DualNumber(0.5,Reverse=True)
    assert EF.Sqrt(Test_Dual_Number_1).val == np.sqrt(0.5) and EF.Sqrt(Test_Dual_Number_1).der == 0

def rm_test_sinh():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Sinh(Test_Dual_Number_1).val == np.sinh(1) and EF.Sinh(Test_Dual_Number_1).der == 0

def rm_test_cosh():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Cosh(Test_Dual_Number_1).val == np.cosh(1) and EF.Cosh(Test_Dual_Number_1).der == 0

def rm_test_tanh():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Tanh(Test_Dual_Number_1).val == np.tanh(1) and EF.Tanh(Test_Dual_Number_1).der == 0
def rm_test_Logistic():
    Test_Dual_Number_1 = DualNumber(1,Reverse=True)
    assert EF.Logistic(Test_Dual_Number_1).val == 1 / (1 + np.exp(-1)) and EF.Logistic(Test_Dual_Number_1).der == 0

rm_test_overload_add()
rm_test_overload_add_types()
rm_test_overload_multiply()
rm_test_overload_multiply_types()
rm_test_overload_sub()
rm_test_overload_sub_types()
rm_test_overload_truediv()
rm_test_overload_truediv_types()
rm_test_overload_power()
rm_test_overload_power_types()
rm_test_overload_pos()
rm_test_overload_neg()
rm_test_overload_abs()
rm_test_overload_round()
rm_test_repr()
rm_test_str()
rm_test_set_der()
rm_test_set_val()
rm_test_jacobian()
rm_test_overload_eq()
rm_test_overload_ne()
rm_test_overload_lt()
rm_test_overload_le()
rm_test_overload_gt()
rm_test_overload_ge()

    
rm_test_sin()      
rm_test_cos()
rm_test_tan()
rm_test_exp()
rm_test_Power()
rm_test_Log()
rm_test_ArcSin()
rm_test_ArcCos()
rm_test_ArcTan()
rm_test_Sqrt()
rm_test_sinh()      
rm_test_cosh()
rm_test_tanh()
rm_test_Logistic()



# Parallelized tests
def pad_test_Jacobian():
    var_names = ['x','y']
    func = ['_x + _y + sin(_x)', 'exp(_x+_y) - _x - sqrt(_y)']
    PAD = Parallelized_AD(fun=func, var=var_names)
    PAD.get_Jacobian([1, 2])
    out = PAD.Jacobian
    assert out[0][0] == 1 + np.cos(1)
    assert out[0][1] == 1
    assert out[1][0] == np.exp(3) - 1
    assert out[1][1] == np.exp(3) - 1 / (2*np.sqrt(2))

def pad_test_value():
    var_names = ['x','y']
    func = ['_x + _y + sin(_x)', 'exp(_x+_y) - _x - sqrt(_y)']
    PAD = Parallelized_AD(fun=func, var=var_names)
    PAD.get_value([1, 2])
    out = PAD.value
    assert out[0] == 1 + 2 + np.sin(1)
    assert out[1] == np.exp(3) - 1 - np.sqrt(2)
     
def pad_test_add_function():
    var_names = ['x','y']
    PAD = Parallelized_AD(var=var_names)
    PAD.add_function('_x + _y + sin(_x)')
    PAD.add_function('exp(_x+_y) - _x - sqrt(_y)')
    PAD.add_function('_x+1+_y ')
    out = PAD.get_Jacobian([1, 2])
    assert out[0][0] == 1 + np.cos(1)
    assert out[0][1] == 1
    assert out[1][0] == np.exp(3) - 1
    assert out[1][1] == np.exp(3) - 1 / (2*np.sqrt(2))
    assert out[2][0] == 1
    assert out[2][1] == 1
    
def pad_test_add_var():
    func = ['_x + _y + sin(_x)', 'exp(_x+_y) - _x - sqrt(_y)']
    PAD = Parallelized_AD(fun=func)
    PAD.add_var('x')
    PAD.add_var('y')
    PAD.add_function('_z**2')
    PAD.add_var('z')
    out = PAD.get_Jacobian([1, 2, 3])
    assert out[0][0] == 1 + np.cos(1)
    assert out[0][1] == 1
    assert out[0][2] == 0
    assert out[1][0] == np.exp(3) - 1
    assert out[1][1] == np.exp(3) - 1 / (2*np.sqrt(2))
    assert out[1][2] == 0
    assert out[2][0] == 0
    assert out[2][1] == 0
    assert out[2][2] == 6
    
def pad_test_get_jacobian():
    var_names = ['x','y']
    func = ['_x + _y + sin(_x)', 'exp(_x+_y) - _x - sqrt(_y)']
    PAD = Parallelized_AD(fun=func, var=var_names)
    out = PAD.get_Jacobian([1, 2])
    assert out[0][0] == 1 + np.cos(1)
    assert out[0][1] == 1
    assert out[1][0] == np.exp(3) - 1
    assert out[1][1] == np.exp(3) - 1 / (2*np.sqrt(2))
    
    out2 = PAD.get_Jacobian([1, 2],forward=True)
    assert out2[0][0] == 1 + np.cos(1)
    assert out2[0][1] == 1
    assert out2[1][0] == np.exp(3) - 1
    assert out2[1][1] == np.exp(3) - 1 / (2*np.sqrt(2))
    
    
def pad_test_get_value():
    var_names = ['x','y']
    func = ['_x + _y + sin(_x)', 'exp(_x+_y) - _x - sqrt(_y)']
    PAD = Parallelized_AD(fun=func, var=var_names)
    out = PAD.get_value([1, 2])
    assert out[0] == 1 + 2 + np.sin(1)
    assert out[1] == np.exp(3) - 1 - np.sqrt(2)
    
def pad_test_forward_reverse():
    '''
    Here's a quick check to ensure our equality condition is working (and our forward
    and reverse modes!).  If everything is in order, forward mode and reverse mode should
    be equal.
    '''
    
    func = ['_x + sin(_y)*_z', '_x + sin(_y)*exp(_z)']
    PAD = Parallelized_AD(fun = func, var = ['x', 'y', 'z'])
    a = PAD.get_Jacobian([1,2,3])
    b = PAD.get_value([1,2,3])
    
    PAD2 = Parallelized_AD(fun = func, var = ['x', 'y', 'z'])
    a2 = PAD2.get_Jacobian([1,2,3], forward=True)
    b2 = PAD2.get_value([1,2,3])

    for i in range(2):
        for j in range(3):
            assert a[i][j] == a2[i][j]
    for i in range(2):
        assert b[i] == b2[i]


pad_test_Jacobian()
pad_test_value()
pad_test_add_function()
pad_test_add_var()
pad_test_get_jacobian()
pad_test_get_value()
pad_test_forward_reverse()



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-

from AD.DualNumber import DualNumber
import numpy as np

def Sin(x):
    '''
    >>> print(Sin(DualNumber(5,1)))
    Derivative: 0.28
    Value: -0.96
    >>> print(Sin(DualNumber(5,Reverse=True)))
    Derivative: 0.00
    Value: -0.96
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.sin(x._val),Reverse=True)
            x.children.append((np.cos(x._val),z))
            return z
        return DualNumber(np.sin(x._val),np.cos(x._val)*x._der)
    else:
        return DualNumber(np.sin(x),0)

    
def Tan(x):
    '''
    >>> print(Tan(DualNumber(5,1)))
    Derivative: 12.43
    Value: -3.38
    >>> print(Tan(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: -3.38
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.tan(x._val),Reverse=True)
            x.children.append(((1+np.tan(x._val)*np.tan(x._val)),z))
            return z
        return DualNumber(np.tan(x._val),(1+np.tan(x._val)*np.tan(x._val))*x._der)
    else:
        return DualNumber(np.tan(x),0)


def Cos(x):
    '''
    >>> print(Cos(DualNumber(5,1)))
    Derivative: 0.96
    Value: 0.28
    >>> print(Cos(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: 0.28
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.cos(x._val),Reverse=True)
            x.children.append((-1*np.sin(x._val),z))
            return z
        return DualNumber(np.cos(x._val),-1*np.sin(x._val)*x._der)
    else:
        return DualNumber(np.cos(x),0)

def Exp(x):
    '''
    >>> print(Exp(DualNumber(5,1)))
    Derivative: 148.41
    Value: 148.41
    >>> print(Exp(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: 148.41
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.exp(x._val),Reverse=True)
            x.children.append((np.exp(x._val),z))
            return z        
        return DualNumber(np.exp(x._val),np.exp(x._val)*x._der)
    else:
        return DualNumber(np.exp(x),0)

def Power(x,n):
    '''
    >>> print(Power(DualNumber(5,1),2))
    Derivative: 10.00
    Value: 25.00
    >>> print(Power(DualNumber(5,Reverse = True),2))
    Derivative: 0.00
    Value: 25.00
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(x._val**n,Reverse=True)
            x.children.append(((n*(x._val**(n-1)),z)))
            return z
        return DualNumber(x._val**n,n*(x._val**(n-1))*x._der)
    else:
        return DualNumber(x**n,0)

def Log(x, base = np.exp(1)):
    '''
    >>> print(Log(DualNumber(5,1)))
    Derivative: 0.20
    Value: 1.61
    >>> print(Log(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: 1.61
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.log(x._val)/np.log(base), Reverse=True)
            x.children.append((1/(x._val*np.log(base)),z))
            return z          
        return DualNumber(np.log(x._val)/np.log(base), (1/(x._val*np.log(base)))*x._der)
    else:
        return DualNumber(np.log(x)/np.log(base), 0)

        
def ArcSin(x):
    '''
    >>> print(ArcSin(DualNumber(0.5)))
    Derivative: 1.15
    Value: 0.52
    >>> print(ArcSin(DualNumber(0.5,Reverse = True)))
    Derivative: 0.00
    Value: 0.52
    '''
    if abs(x) >= 1:
        raise ValueError('ArcSin is only defined on (-1,1)!')
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.arcsin(x._val),Reverse=True)
            x.children.append((1/np.sqrt(1-x._val**2),z))
            return z          
        return DualNumber(np.arcsin(x._val),1/np.sqrt(1-x._val**2) * x._der)
    else:
        return DualNumber(np.arcsin(x),0)


def ArcCos(x):
    '''
    >>> print(ArcCos(DualNumber(0.5)))
    Derivative: -1.15
    Value: 1.05
    >>> print(ArcCos(DualNumber(0.5,Reverse = True)))
    Derivative: 0.00
    Value: 1.05
    '''
    if abs(x) >= 1:
        raise ValueError('ArcCos is only defined on (-1,1)!')
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.arccos(x._val),Reverse=True)
            x.children.append((-1/np.sqrt(1-x._val**2),z))
            return z
        return DualNumber(np.arccos(x._val),-1/np.sqrt(1-x._val**2) * x._der)
    else:
        return DualNumber(np.arccos(x),0)
            
def ArcTan(x):
    '''
    >>> print(ArcTan(DualNumber(0.5)))
    Derivative: 0.80
    Value: 0.46
    >>> print(ArcTan(DualNumber(0.5,Reverse = True)))
    Derivative: 0.00
    Value: 0.46
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.arctan(x._val),Reverse=True)
            x.children.append((1/(1+x._val**2),z))
            return z        
        return DualNumber(np.arctan(x._val),1/(1+x._val**2) * x._der)
    else:
        return DualNumber(np.arctan(x),0)

def Sqrt(x):
    '''
    >>> print(Sqrt(DualNumber(9,1)))
    Derivative: 0.17
    Value: 3.00
    >>> print(Sqrt(DualNumber(9,Reverse = True)))
    Derivative: 0.00
    Value: 3.00
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(np.sqrt(x._val),Reverse=True)
            x.children.append((1/(2*np.sqrt(x._val)),z))
            return z             
        return DualNumber(np.sqrt(x._val),1/(2*np.sqrt(x._val)) * x._der)
    else:
        return DualNumber(np.sqrt(x),0)

# Hyperbolic functions
def Sinh(x):
    '''
    >>> print(Sinh(DualNumber(5,1)))
    Derivative: 74.21
    Value: 74.20
    >>> print(Sinh(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: 74.20
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber((np.exp(x._val)-np.exp(-x._val))/2,Reverse=True)
            x.children.append(((np.exp(x._val)+np.exp(-x._val))/2, z))
            return z
        return DualNumber((np.exp(x._val)-np.exp(-x._val))/2,((np.exp(x._val)+np.exp(-x._val))/2)*x._der)
    else:
        return DualNumber((np.exp(x)-np.exp(-x))/2,0)

def Cosh(x):
    '''
    >>> print(Cosh(DualNumber(5,1)))
    Derivative: 74.20
    Value: 74.21
    >>> print(Cosh(DualNumber(5,Reverse =  True)))
    Derivative: 0.00
    Value: 74.21
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber((np.exp(x._val)+np.exp(-x._val))/2,Reverse=True)
            x.children.append(((np.exp(x._val)-np.exp(-x._val))/2, z))
            return z
        return DualNumber((np.exp(x._val)+np.exp(-x._val))/2,((np.exp(x._val)-np.exp(-x._val))/2)*x._der)
    else:
        return DualNumber((np.exp(x)+np.exp(-x))/2,0)

def Tanh(x):
    '''
    >>> print(Tanh(DualNumber(5,1)))
    Derivative: 0.00
    Value: 1.00
    >>> print(Tanh(DualNumber(5,Reverse =  True)))
    Derivative: 0.00
    Value: 1.00
    '''
    if data_type_check(x) == 0:
        Z = (np.exp(x._val)-np.exp(-x._val))/(np.exp(x._val)+np.exp(-x._val))
        if x._rev:
            z=DualNumber(Z,Reverse=True)
            x.children.append((1-Z**2, z))
            return z
        return DualNumber(Z,(1-Z**2)*x._der)
    else:
        return DualNumber((np.exp(x)-np.exp(-x))/(np.exp(x)+np.exp(-x)), 0)

# Logistic functions
def Logistic(x):
    '''
    >>> print(Logistic(DualNumber(5,1)))
    Derivative: 0.01
    Value: 0.99
    >>> print(Logistic(DualNumber(5,Reverse = True)))
    Derivative: 0.00
    Value: 0.99
    '''
    if data_type_check(x) == 0:
        if x._rev:
            z=DualNumber(1/(1+np.exp(-x._val)),Reverse=True)
            x.children.append((np.exp(x._val)/(1+np.exp(x._val))**2, z))
            return z
        return DualNumber(1/(1+np.exp(-x._val)),(np.exp(x._val)/(1+np.exp(x._val))**2)*x._der)
    else:
        return DualNumber(1/(1+np.exp(-x)), 0)


def data_type_check(x):
   try:
       if x._der==None and x._rev==True:
           return 0  # returns 0 if x is DualNumber
       float(x._val)+float(x._der)
       return 0  # returns 0 if x is DualNumber
   except AttributeError:
       try:
           float(x)
           return 1 # returns 1 if x is real
       except:
           raise AttributeError('Input must be dual number or real number!')


if __name__ =="__main__":
    #x = DualNumber(0.5,Reverse=True)
    #y = DualNumber(4.2,Reverse=True)
    #z = x * y + Sin(x)
    #z.set_der(value=1.0)
    #print('value of x*y + sin(x) evaluated at x=0.5, y=4.2: {}\nOur implementation gives: {}'.format(0.5 * 4.2 + np.sin(0.5), z.val))
    #print('value of dz/dx = y + cos(x) evaluated at x=0.5, y=4.2: {}\nReverse method of our implementation: {}'.format(4.2 + np.cos(0.5), x.der))
    #print('value of dz/dy = x evaluated at x=0.5, y=4.2: {}\nReverse method of our implementation: {}'.format(0.5, y.der))

    import doctest
    doctest.testmod(verbose=True)



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Root_Finding Algorithm
Created on Thu Nov 14 13:56:17 2019
"""

from AD.DualNumber import DualNumber
import AD.ElementaryFunctions as EF

"""
Demo code to find root of an arbitrary function based on our AD framework
"""
def Given_function(x):
    x=DualNumber(x);
    y=EF.Sin(x)*EF.Cos(x)*EF.Tan(x)-EF.Exp(x)*EF.Log(x)*EF.Sqrt(x)*2
    return y #y=sin(x)cos(x)tan(x)-2exp(x)log(x)sqrt(x)
x0=1.2#Initialization of root finding algorithm
Max_iter=int(1e4) #Maximum amout of iterations before stop
Threshold=1e-12;
x=x0
for i in range(Max_iter):
    Current_value=Given_function(x).val
    current_slope=Given_function(x).der
    delta_X=-Current_value/current_slope
    if abs(delta_X)<Threshold:
        break
    else:
        x=x+delta_X
print('The found root of 0=sin(x)cos(x)tan(x)-2exp(x)log(x)sqrt(x) is', x)
print('The numerical solution given by other software is 1.132182161975462')



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 13:20:32 2019

"""

import numpy as np


class DualNumber():

    """Returns DualNumber and defines the overloading of arithmetric and unary operators.
    INPUTS
    =======
    real: float
      This stores the value of the function for which the derivative would be evaluated at.
    dual: float, optional, default value is 1
      This stores the value for the derivative of the function.
    Reverse: Boolean, optional, if set to true the gradient will be calculated based on reverse mode
        Default value is false, forward mode is used
    RETURNS
    ========
    DualNumber: An instance with methods such as val() and der() to get out the current value and derivative of function respectively.
    NOTES
    =====
    PRE:
        - real, dual have numeric types
        - input of real is needed but input of real is optional
    POST:
        - real will be stored in the private attribute self._val and holds the current value of the function
        - dual will be stored in the private attribute self._der and holds the current derivative value of the function
        - raises a AttributeError exception if either real or dual is not of numeric type
        - returns a DualNumber instance
        - defines the overloading of arithmetic operators and unary operators
    EXAMPLES
    =========
    >>> [DualNumber(3,2).val, DualNumber(3,2).der]
    [3, 2]
    >>> [DualNumber(2.3,2).val, DualNumber(2.3,2).der]
    [2.3, 2]
    >>> [DualNumber(2,2.3).val, DualNumber(2,2.3).der]
    [2, 2.3]
    >>> [DualNumber(2,Reverse=True).val, DualNumber(2,Reverse=True).der]
    [2, 0]
    >>> [DualNumber(2.3,2.3).val, DualNumber(2.3,2.3).der]
    [2.3, 2.3]
    >>> print(DualNumber(2.3,2.3))
    Derivative: 2.30
    Value: 2.30
    """

    def __init__(self, real, dual=1,Reverse=False):
        #print(real)
        assert (isinstance(real, int) or isinstance(real, float)), "Check the type of real!"
        assert (isinstance(dual, float) or isinstance(dual, int)), "Check the type of dual!"
        self._val = real
        self._der = dual
        self._rev = Reverse
        if Reverse is True:
            self._der = None
            self.children = []
            self._rev = True #Working on reverse mode
    @property
    def val(self):
        return self._val
    
    @property
    def der(self):
        if self._der is None:
            self._der = sum(weight * var.der for weight, var in self.children)
        return self._der    
    @property
    def jacobian(self):
        return self.der
    
    def __repr__(self):
        return 'Derivative: {0:.2f}'.format(self.der) + '\n' + 'Value: {0:.2f}'.format(self.val)
        
    def __str__(self):
        return('Derivative: {0:.2f}'.format(self.der) + '\n' + 'Value: {0:.2f}'.format(self.val))
    def set_der(self,value=None):
        #Set derivative of current dual number
        if value: self._der=value        
    def set_val(self,value=None):
        #Set val of current dual number
        if value: self._val=value        
# Overloading arithmetic operators

    def __add__(self, other):
        '''
        >>> print(DualNumber(5,1)+3)
        Derivative: 1.00
        Value: 8.00
        >>> Z =DualNumber(5,Reverse=True)+3
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 8.00
        '''
        try:
            if self._rev:
                z = DualNumber(self._val + other._val,Reverse=True)
                self.children.append((1.0, z))
                other.children.append((1.0, z))     
                return z
            else:
                val2 = self.val + other.val
                der2 = self.der + other.der
                return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(self._val + other,Reverse=True)
                self.children.append((1.0, z))
                return z                    
            else:
                return DualNumber(self.val + other, self.der)

    def __radd__(self, other):
        '''
        >>> print(3+DualNumber(5,1))
        Derivative: 1.00
        Value: 8.00
        >>> Z =3+DualNumber(5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 8.00
        '''
        return self.__add__(other)

    def __mul__(self, other):
        '''
        >>> print(DualNumber(5,1)*3)
        Derivative: 3.00
        Value: 15.00
        >>> Z =DualNumber(5,Reverse=True)*3
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 15.00
        '''
        try:
            if self._rev:
                z = DualNumber(self._val * other._val,Reverse=True)
                self.children.append((other.val, z))
                other.children.append((self.val, z)) 
                return z
            else:
                val2 = self.val * other.val
                der2 = self.der * other.val + self.val * other.der
                return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(self._val * other,Reverse=True)
                self.children.append((other, z))
                return z
            val2 = self.val * other
            der2 = self.der * other
            return DualNumber(val2, der2)

    def __rmul__(self, other):
        '''
        >>> print(3*DualNumber(5,1))
        Derivative: 3.00
        Value: 15.00
        >>> Z = 3*DualNumber(5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 15.00
        '''
        return self.__mul__(other)

    def __sub__(self, other):
        '''
        >>> print(DualNumber(5,1)-3)
        Derivative: 1.00
        Value: 2.00
        >>> Z =DualNumber(5,Reverse=True)-3
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 2.00
        '''
        try:
            if self._rev:
                z = DualNumber(self.val - other.val,Reverse=True)
                self.children.append((1, z))
                other.children.append((-1, z))
                return z  
            else:
                val2 = self.val - other.val
                der2 = self.der - other.der
                return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(self._val - other,Reverse=True)
                self.children.append((1, z))
                return z 
            else:
                val2 = self.val - other
                der2 = self.der
                return DualNumber(val2, der2)

    def __rsub__(self, other):
        '''
        >>> print(3-DualNumber(5,1))
        Derivative: -1.00
        Value: -2.00
        >>> Z = 3 - DualNumber(5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: -2.00
        '''
        try:
            if self._rev:
                z = DualNumber(other.val - self.val,Reverse=True)
                self.children.append((-1, z))
                other.children.append((1, z))
                return z                
            else:
                val2 = other.val - self.val
                der2 = other.der - self.der
                return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(other - self.val,Reverse=True)
                self.children.append((-1, z))
                return z
            else:
                val2 = other - self.val
                der2 = -self.der
                return DualNumber(val2, der2)

    def __truediv__(self, other):
        '''
        >>> print(DualNumber(5,1)/2)
        Derivative: 0.50
        Value: 2.50
        >>> Z = DualNumber(5,Reverse=True)/2
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 2.50
        '''
        try:
            if self._rev:
                z = DualNumber(self.val / other.val,Reverse=True)
                self.children.append((1/other.val , z))
                other.children.append((-self.val/(other.val*other.val) , z))
                return z 
            else:                 
                val2 = self.val / other.val
                der2 = (-self.val * other.der + self.der*other.val)/(other.val*other.val)
                return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(self.val / other,Reverse=True)
                self.children.append((1/other , z))
                return z
            else:
                val2 = self.val / other
                der2 = self.der / other
                return DualNumber(val2, der2)

    def __rtruediv__(self, other):
        '''
        >>> print(2/DualNumber(5,1))
        Derivative: -0.08
        Value: 0.40
        >>> Z = 2/DualNumber(5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 0.40
        '''
        try:
            if self._rev:
                z = DualNumber(other.val/self.val,Reverse=True)
                self.children.append((-other.val/self.val/self.val , z))
                other.children.append((1/self.val , z))
                return z            
            else:
                val2 = other.val / self.val
                der2 = (other.der * self.val - other.val*self.der)/(self.val*self.val)
                return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(other/self.val,Reverse=True)
                self.children.append((-other/self.val/self.val , z))
                return z             
            else:
                val2 = other / self.val
                der2 = -other*self.der / (self.val*self.val)
                return DualNumber(val2, der2)

    def __pow__(self, other):
        '''
        >>> print(DualNumber(5,1)**2)
        Derivative: 10.00
        Value: 25.00
        >>> Z = DualNumber(5,Reverse=True)**2
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 25.00
        '''
        try:
            if self._rev:
                z = DualNumber(self.val ** other.val,Reverse=True)
                self.children.append(((self.val ** other.val)*other.val/self.val , z))
                other.children.append(((self.val ** other.val)*np.log(self.val), z))
                return z      
            else:
                val2 = self.val ** other.val
                der2 = val2*(other.val/self.val*self.der+other.der*np.log(self.val))
                return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if self._rev:
                z = DualNumber(self.val ** other,Reverse=True)
                self.children.append(((self.val ** other)*other/self.val , z))
                return z         
            else:
                val2 = self.val ** other
                der2 = other * (self.val ** (other - 1)) * self.der
                return DualNumber(val2, der2)


    def __rpow__(self, other):
        '''
        >>> print(2**DualNumber(5,1))
        Derivative: 22.18
        Value: 32.00
        >>> Z = 2**DualNumber(5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 32.00
        '''
        try:
            if self._rev:
                z = DualNumber(other.val ** self.val,Reverse=True)
                other.children.append(((other.val ** self.val)*self.val/other.val , z))
                self.children.append(((other.val ** self.val)*np.log(other.val), z))
                return z   
            else:
                if other.val <= 0:
                    raise AttributeError('Base must be >0!')
                val2 = other.val ** self.val
                der2 = val2*(self.val/other.val*other.der+self.der*np.log(other.val))
                return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            if other <= 0:
                    raise AttributeError('Base must be >0!')
            if self._rev:
                z = DualNumber(other ** self.val,Reverse=True)
                self.children.append((other ** self.val * np.log(other), z))
                return z   
            else:
                val2 = other ** self.val
                der2 = other ** self.val * np.log(other) * self.der
                return DualNumber(val2, der2)

# Overloading unary operators
    def __pos__(self):
        '''
        >>> print(+DualNumber(-5,-1))
        Derivative: -1.00
        Value: -5.00
        >>> Z = +DualNumber(-5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: -5.00
        '''
        if self._rev:
            z = DualNumber(self.val,Reverse=True)
            self.children.append((1, z))
            return z
        else:
            val2 = self.val
            der2 = self.der
            return DualNumber(val2, der2)

    def __neg__(self):
        '''
        >>> print(-DualNumber(-5,-1))
        Derivative: 1.00
        Value: 5.00
         >>> Z = -DualNumber(-5,Reverse=True)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 5.00
        '''
        if self._rev:
            z = DualNumber(-self.val,Reverse=True)
            self.children.append((-1, z))
            return z    
        else:
            val2 = -self.val
            der2 = -self.der
            return DualNumber(val2, der2)

    def __abs__(self):
        '''
        >>> print(abs(DualNumber(-5,-1)))
        Derivative: 1.00
        Value: 5.00
        >>> Z = abs(DualNumber(-5,Reverse=True))
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: 5.00
        '''
        if self._rev:
            z = DualNumber(abs(self.val),Reverse=True)
            if self.val >= 0:
                self.children.append(( 1, z))
            else:
                self.children.append((-1, z))
            return z                
        else:
            val2 = abs(self.val)
            if self.val >= 0:
                der2 = self.der
            else:
                der2 = -1 * self.der
            return DualNumber(val2, der2)

    def __round__(self, n=None):
        '''
        >>> print(round(DualNumber(-5.234,-1.256),2))
        Derivative: -1.26
        Value: -5.23
        >>> Z = round(DualNumber(-5.234,Reverse=True), 2)
        >>> Z.set_der(1)
        >>> print(Z)
        Derivative: 1.00
        Value: -5.23
        '''
        if self._rev:
            z = DualNumber(round(self.val, n),Reverse=True)
            self.children.append((1, z))
            return z    
        else:
            val2 = round(self.val, n)
            der2 = round(self.der, n)
            return DualNumber(val2, der2)
        
     
    # Overloading comparison operators
    def __eq__(self, other):
        '''
        >>> DualNumber(-5,-1)==DualNumber(-5,-1)
        True
        >>> DualNumber(-5,Reverse=True)==DualNumber(-5,Reverse=True)
        True
        '''
        assert isinstance(other,DualNumber)
        if not self._rev and not other._rev and self.val == other.val and self.der == other.der :
            return True
        if self._rev and other._rev and self.val == other.val:
            return True
        return False

    def __ne__(self, other):
        '''
        >>> DualNumber(-5,-1)!=DualNumber(-5,-2)
        True
        >>> DualNumber(-5,Reverse=True)!=DualNumber(-5,Reverse=True)
        False
        '''
        assert isinstance(other, DualNumber)
        if not self._rev and not other._rev and self.val == other.val and self.der == other.der :
            return False
        if self._rev and other._rev and self.val == other.val:
            return False
        return True

    def __lt__(self, other):
        '''
        >>> DualNumber(-5,-1) < DualNumber(-4,-2)
        True
        >>> DualNumber(-5,Reverse=True) < DualNumber(-4,Reverse=True)
        True
        '''
        assert (isinstance(other, DualNumber) or isinstance(other, int) or isinstance(other, float))
        if isinstance(other,DualNumber):
            if self.val < other.val:
                return True
        if isinstance(other, int) or isinstance(other, float):
            if self.val < other:
                return True
        
        return False

    def __le__(self, other):
        '''
        >>> DualNumber(-5,-1) <= DualNumber(-5,-2)
        True
        >>> DualNumber(-5,Reverse=True) <= DualNumber(-5,Reverse=True)
        True
        '''
        assert (isinstance(other, DualNumber) or isinstance(other, int) or isinstance(other, float))
        if isinstance(other,DualNumber):
            if self.val <= other.val:
                return True
        if isinstance(other, int) or isinstance(other, float):
            if self.val <= other:
                return True
            
        return False

    def __ge__(self, other):
        '''
        >>> DualNumber(-5,-1) >= -6
        True
        >>> DualNumber(-5,Reverse=True) >= DualNumber(-5,Reverse=True)
        True
        '''
        assert (isinstance(other, DualNumber) or isinstance(other, int) or isinstance(other, float))
        if isinstance(other,DualNumber):
            if self.val >= other.val:
                return True
        if isinstance(other, int) or isinstance(other, float):
            if self.val >= other:
                return True
            
        return False

    def __gt__(self, other):
        '''
        >>> DualNumber(5,-1) > DualNumber(-5,-2)
        True
        >>> DualNumber(5,Reverse=True) > DualNumber(-5,Reverse=True)
        True
        '''
        assert (isinstance(other, DualNumber) or isinstance(other, int) or isinstance(other, float))
        if isinstance(other,DualNumber):
            if self.val > other.val:
                return True
        elif isinstance(other, int) or isinstance(other, float):
            if self.val > other:
                return True
            
        return False
if __name__ =="__main__":
    # x=DualNumber(-2.578,-1.2345)
    # y=DualNumber(3,1)
    # print(x)
    # f=round(x,2)
    # print(f.val,f.der)
    import doctest
    doctest.testmod()
    

    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Dec  2 01:26:47 2019
Adams Optimization Algorithm based on
https://arxiv.org/abs/1412.6980
Implemented based on AutoDiffing library
CS207 Final Project
"""

from AD.DualNumber import DualNumber
import AD.ElementaryFunctions as EF
import numpy as np
class Adam_optimizer():
    def __init__(self,Objective_Grad=None,Target_Var=None):
        '''
        Objective_Grad: Gradient of Objective Function to be optimized
        Target_Var: Target variables need to be optimized
        '''
        self.Objective_Grad=Objective_Grad
        self.Target_Var=Target_Var
        self.set_parameters()
    def set_parameters(self,alpha=0.01,beta1=0.9,beta2=0.999,epsilon=1e-8,max_iter=1e5,tol=1e-8):
        '''
        Setting up hyderparameters used for the optimizer
        Adam has four hyperparameters:
            beta_1:float from 0 to 1, usually close to 1
            beta_2:float from 0 to 1, usually close to 1
            alpha :float, step size, usually close to zero
            epsilon :float, usually very close to zero
            max_iter :interger, max interations before force stop
            tol :float, tolerance before the optimzer should stop
        Detailed definition please refer to https://arxiv.org/pdf/1412.6980.pdf
        Default values borrowed from original paper, max iter set to a hundred thousand.
        '''
        self.alpha=alpha
        self.beta_1=beta1
        self.beta_2=beta2
        self.epsilon=epsilon
        self.max_iter=max_iter
        self.tol=tol

    def set_objective(self,Objective_Grad):
        '''
        Setting up objective function gradient
        input:
            Objective_Grad: function, A function returns the gradient of objective 
        
        '''
        self.Objective_Grad=Objective_Grad
#    def set_target_variables(self,Target_Var):
#        '''
#        Setting up objective function
#        input:
#            Target_Var: A list of variables, 
#        
#        '''
#        self.Target_Var=Target_Var
    def optimize(self,initialization=1):
        self._t=0 # Time =0
        self._m=0 # Initialize first momentum
        self._v=0# Initialize second momentum
        self._theta=initialization
        self.log=[]
        for i in np.arange(self.max_iter):
        	self._t+=1
        	g_t = self.Objective_Grad(self._theta)		#Get gradients w.r.t. stochastic objective at timestep t
        	self._m = self.beta_1*self._m + (1-self.beta_1)*g_t	#Update biased first moment estimate
        	self._v = self.beta_2*self._v + (1-self.beta_2)*(g_t*g_t)	#Update biased second raw moment estimate)
        	m_cap = self._m/(1-(self.beta_1**self._t))		#Compute bias-corrected first moment estimate
        	v_cap = self._v/(1-(self.beta_2**self._t))		#Compute bias-corrected second raw moment estimate
        	theta1 = self._theta - (self.alpha*m_cap)/(np.sqrt(v_cap)+self.epsilon)	#Update parameters
        	if abs( theta1-self._theta)< self.tol:		#checks convergence
        		break   
        	self._theta = theta1
        	self.log.append(theta1)
        return self._theta

if __name__=='__main__':
    def Demo_gradient(x):
        x=DualNumber(x);
        y=-EF.Sin(x)*EF.Cos(x)*EF.Tan(x)+EF.Exp(x)*EF.Log(x)*EF.Sqrt(x)*2
        return y.der #y=sin(x)cos(x)tan(x)-2exp(x)log(x)sqrt(x)
    Ad=Adam_optimizer(Demo_gradient)
    optimized_location=Ad.optimize()
    print('y=-sin(x)cos(x)tan(x)+2exp(x)log(x)sqrt(x) optimized starting from intialization 1 yields:\n',optimized_location)
    
    # import matplotlib.pyplot as plt
    # x= np.linspace(0.01, 1.2,1000)
    # y=-np.sin(x)*np.cos(x)*np.tan(x)+2*np.exp(x)*np.log(x)*np.sqrt(x)
    # plt.figure(dpi=500)
    # plt.plot(x,y,label='Objective')
    # x=Ad.log
    # y=-np.sin(x)*np.cos(x)*np.tan(x)+2*np.exp(x)*np.log(x)*np.sqrt(x)
    # plt.plot(x,y,'x',label='Trace')
    # plt.xlabel('X')
    # plt.ylabel('Y')
    # plt.legend()
    # plt.title('Trajectory of Adam')
    
        